/*---------------------------------------------------------------------------
  This code is part of the book
  LOGIC PROGRAMMING TECHNIQUES: THE PROLOG LANGUAGE
  Ilias Sakellariou, Nikos Bassiliades, Petros Kefalas, Demosthenes Stamatis
  Hellenic Academic On-line Publications, 2015
----------------------------------------------------------------------------*/

/*------------------------------------------------------------------
               COMPILER FINITE STATE MACHINE LANGUAGE
-------------------------------------------------------------------*/

:-[fsmllexical].

:-dynamic(errors/1).


compile:-
	writelist([nl,'   ************** FSML Compiler *****************',nl,nl]),
	writelist(['Input Filename and extention (in single quotes) :']),
	read(IF),
	writelist(['Compiling ', IF, '......', nl,nl]),
	initialise_errors,
	writelist(['Target Filename (.pl) :']),
	read(OF),
	see(IF),
	tell(OF),
	writelist([':-[fsmlanimator].', nl,nl]),
	compile_code,
	told,
	report_errors,!.

compile_code:-
	repeat,
	myread(Statement,S),
	(S = ['#end'] -> 
			(seen,!);
			(parse(Statement,S,[]),
			fail)).


/*-------------------------------------------------------------------
  ALL LANGUAGE CONSTRUCTS AND THE CORRESPONDING TRANSLATIONS
--------------------------------------------------------------------*/

parse(_,S,[]):- model(ModelName,S,[]),!,
		translate(fsm(ModelName)),!.
parse(_,S,[]):- init_state(InitState,S,[]),!,
		translate(initial_state(InitState)).
parse(_,S,[]):- final_states(FinalStates,S,[]),!,
		translate(final_states(FinalStates)).
parse(_,S,[]):- states(StateList,S,[]),!,
		translate(states(StateList)).
parse(_,S,[]):- transition(S1,F,S2,S,[]),!,
		translate(transition(S1,F,S2)).
parse(_,S,[]):- input(I,S,[]),!,
		translate(input(I)).
parse(_,S,[]):- accept(Query,S,[]),!,
		translate(Query).
parse(L,_,[]):-	
		writelist([nl,'****** Syntax Error in ******',nl]),
		write(L),nl,
		writelist(['*****************************',nl,nl]),
		retract(errors(X)), X1 is X+1, assert(errors(X1)),!.

translate(accept(Q)):-!,
	writelist([':-',accept(Q),'.',nl]),!.

translate(X):-
	write(X),write('.'),nl,!.


/*-------------------------------------------------------------------
                           MODEL NAME
--------------------------------------------------------------------*/
model(ModelName) --> 
	['#fsm'], 
	constant(ModelName).

/*-------------------------------------------------------------------
                         STATES
--------------------------------------------------------------------*/
states(States)--> 	
	['#states'],
	['='],
	set(States).
	
/*-------------------------------------------------------------------
                         INITIAL STATE
--------------------------------------------------------------------*/
init_state(InitState)--> 	
	['#initial_state'],
	['='],
	constant(InitState).

/*-------------------------------------------------------------------
                         FINAL STATES
--------------------------------------------------------------------*/
final_states(FinalStates)--> 	
	['#final_states'],
	['='],
	set(FinalStates).

/*-------------------------------------------------------------------
                         INPUTS
--------------------------------------------------------------------*/
input(InputSet)--> 	
	['#input'],
	['='],
	set(InputSet).


/*-------------------------------------------------------------------
                    TRANSITIONS BETWEEN STATES
--------------------------------------------------------------------*/
transition(S1,I,S2) --> 
	['#transition'],
	['from'],
	constant(S1),
	['with'],
	constant(I),
	['to'],
	constant(S2).


/*-------------------------------------------------------------------
                         QUERY
--------------------------------------------------------------------*/
accept(accept(QueryParameter))--> 	
	['#accept'],
	sequence(QueryParameter).

/*-------------------------------------------------------------------
                             SETS & SEQUENCES
--------------------------------------------------------------------*/

set(L) -->
	['{'], 
	set_elements(L), 
	['}'].

sequence(S) -->
	['<'], 
	set_elements(S), 
	['>'].

set_elements([Elem])-->
	constant(Elem).
set_elements([Head|Tail])-->
	constant(Head),
	[','],
	set_elements(Tail).


/*-------------------------------------------------------------------
           CONSTANTS
--------------------------------------------------------------------*/

constant(WC)-->	constant2(L),		{wordtolist(WC,L)}.


constant2([C]) -->
	allowedchar(C).
constant2([C|R]) --> 	
	allowedchar(C), 
	constant2(R).

/*-------------------------------------------------------------------
                 SYMBOLS ALLOWED IN IDENTIFIERS
--------------------------------------------------------------------*/
allowedchar(L)-->	letter(L).
allowedchar(D)-->	digit(D).
allowedchar(S)-->	special_symbol(S).

letter(X)-->		[X],
	{member(X,[a,b,c,d,e,f,g,h,i,j,k,l,m,n,o,p,q,r,s,t,u,v,w,x,y,z])}.

digit(X)-->		[X],
	{member(X,[0,1,2,3,4,5,6,7,8,9])}.

special_symbol('_')-->	['_'].


/*-------------------------------------------------------------------
  AUXILIARY PREDICATES
--------------------------------------------------------------------*/

initialise_errors:-
	abolish(errors/1),
	assert(errors(0)),!.

report_errors:-
	errors(E),!,
	writelist([nl, E, ' Errors found!',nl,nl]),!.
	

writelist([]):-!.
writelist([tab|T]):-!,
	tab(5),
	writelist(T).
writelist([nl|T]):-!,
	nl,
	writelist(T).
writelist([H|T]):-!,
	write(H),
	writelist(T).
	
/*-------------------------------------------------------------------
  SYMBOL TABLE
--------------------------------------------------------------------*/	

tag('#fsm').
tag('#states').
tag('#input').
tag('#initial_state').
tag('#final_states').
tag('#transition').
tag('#accept').
tag('from').
tag('with').
tag('to').
tag('#end').